package com.sun.org.apache.bcel.internal.classfile;

import java.util.Objects;
import java.io.DataInput;
import com.sun.org.apache.bcel.internal.Const;
import java.io.IOException;
import java.io.DataOutputStream;
import com.sun.org.apache.bcel.internal.util.BCELComparator;

public abstract class Constant implements Cloneable, Node
{
    private static BCELComparator bcelComparator;
    private byte tag;
    
    Constant(byte tag) {
        this.tag = tag;
    }
    
    @Override
    public abstract void accept(Visitor p0);
    
    public abstract void dump(DataOutputStream p0) throws IOException;
    
    public final byte getTag() {
        return this.tag;
    }
    
    @Override
    public String toString() {
        return Const.getConstantName((int)this.tag) + "[" + (int)this.tag + "]";
    }
    
    public Constant copy() {
        try {
            return (Constant)super.clone();
        }
        catch (CloneNotSupportedException ex) {
            return null;
        }
    }
    
    public Object clone() {
        try {
            return super.clone();
        }
        catch (CloneNotSupportedException e) {
            throw new Error("Clone Not Supported");
        }
    }
    
    public static Constant readConstant(DataInput dataInput) throws IOException, ClassFormatException {
        byte b = dataInput.readByte();
        switch (b) {
            case 7: {
                return new ConstantClass(dataInput);
            }
            case 9: {
                return new ConstantFieldref(dataInput);
            }
            case 10: {
                return new ConstantMethodref(dataInput);
            }
            case 11: {
                return new ConstantInterfaceMethodref(dataInput);
            }
            case 8: {
                return new ConstantString(dataInput);
            }
            case 3: {
                return new ConstantInteger(dataInput);
            }
            case 4: {
                return new ConstantFloat(dataInput);
            }
            case 5: {
                return new ConstantLong(dataInput);
            }
            case 6: {
                return new ConstantDouble(dataInput);
            }
            case 12: {
                return new ConstantNameAndType(dataInput);
            }
            case 1: {
                return ConstantUtf8.getInstance(dataInput);
            }
            case 15: {
                return new ConstantMethodHandle(dataInput);
            }
            case 16: {
                return new ConstantMethodType(dataInput);
            }
            case 17: {
                return new ConstantDynamic(dataInput);
            }
            case 18: {
                return new ConstantInvokeDynamic(dataInput);
            }
            case 19: {
                return new ConstantModule(dataInput);
            }
            case 20: {
                return new ConstantPackage(dataInput);
            }
            default: {
                throw new ClassFormatException("Invalid byte tag in constant pool: " + (int)b);
            }
        }
    }
    
    public static BCELComparator getComparator() {
        return Constant.bcelComparator;
    }
    
    public static void setComparator(BCELComparator comparator) {
        Constant.bcelComparator = comparator;
    }
    
    @Override
    public boolean equals(Object obj) {
        return Constant.bcelComparator.equals(this, obj);
    }
    
    @Override
    public int hashCode() {
        return Constant.bcelComparator.hashCode(this);
    }
    
    static {
        Constant.bcelComparator = new BCELComparator() {
            @Override
            public boolean equals(Object o1, Object o2) {
                Constant THIS = (Constant)o1;
                Constant THAT = (Constant)o2;
                return Objects.equals(THIS.toString(), THAT.toString());
            }
            
            @Override
            public int hashCode(Object o) {
                Constant THIS = (Constant)o;
                return THIS.toString().hashCode();
            }
        };
    }
}
